/* gts2pov.c
 * convert GNU Triangulated Surfaces (GTS) data to PovRay mesh[2].
 */

/* setreuid(2), strdup(3) */
#define _XOPEN_SOURCE 500

#include <assert.h>
#include <errno.h>
#include <locale.h>
#include <math.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#define G_DISABLE_DEPRECATED  1
#include <gts.h>

/* arbitrary limit on vertices */
#ifndef VERTEXMAX
#define VERTEXMAX 1000000
#endif

/* conveniences + misc constants */
#define DFLTIDENT    "mesh"
#define DFLTPREFIX   "gts_"
#define DFLTVERSION  3.8

#define IDENTMAX  40

#define AUXLNMAX  2048

#define CHRNUL  '\0'

#define AUU  __attribute__((unused))

#define EEXIT  (eExit(__func__,__LINE__))

/* file global.  CLI stuff, mostly */
static char const * const prgname = "gts2pov";
static char const * const prgversion = "0.3.0";
/* -1       mesh.            default 'mesh2'.
   -a file  auxiliary.       default none.
   -b       "bare", no include wrapper.
   -D       diagnostic stats.
   -f       force.
   -h       help.
   -i axis  inside vector.   default none.
   -n name  identifier.      default 'mesh'.
   -P n.n   povray version.  default '3.8'.
   -p abc   common prefix.   default 'gts_'.
   -s       smoothed triangles.
   -[Vv]    version.
 */
static char const * const prgopts = "+1a:bDfhi:n:P:p:sVv";
/* only one message for '-[Vv]', '-D' undocumented */
static char const * const prgopts_help[] = {
  "1\t\tgenerate older 'mesh' object instead",
  "a filename\t'[object modifiers]' to insert",
  "b\t\tgenerate \"bare\" object, ie no .inc wrap",
  "f\t\tforce, when mesh not closed",
  "h\t\tdisplay this message",
  "i x|y|z\tinside vector for mesh",
  "n name\tmesh name/identifier",
  "P n.n\tactive POV-Ray version",
  "p name\tidentifier prefix",
  "s\t\tgenerate smoothed triangles",
  "V\t\tdisplay the program version",
  NULL
};

enum InsideVect {kNone = 0, kX, kY, kZ};
typedef enum InsideVect InsideVect;

static char *aux_value;
static FILE *aux_fp;
static bool diag;
static bool mkbare;
static bool mkforce;
static bool mkmesh2;
static bool mksmooth;
static double povvers_value;
static char *prefix_value;
static char *ident_value;
static InsideVect inside;

static char const * const ivecmask = "  inside_vector %s\n",
                  * const auxmask =
        "/* ----------<auxiliary %s '%s'>---------- */\n";

static char const * const inside_vectors [] = {
  "", "<1,0,0>", "<0,1,0>", "<0,0,1>", NULL
};

typedef struct Cdata Cdata;
struct Cdata {
  GtsSurface * const s;
  gdouble const diff;
};

typedef struct Vector Vector;
struct Vector {
  gdouble x, y, z;
  bool done;
};
static Vector *fnorms, *vnorms;

static gdouble arealo, areahi;

/* ------------------------------------------------------------------------- */

static void eExit(char const * const, int const) __attribute__((noreturn));
static void eExit(char const * const fn, int const ln) {
  int const e = errno;
  fprintf(stderr, "%s: oops in '%s' line %d, error (%d) %s.\n",
          prgname, fn, ln, e, strerror(e));
  exit(EXIT_FAILURE);
}

static void usageHelp(int const) __attribute__((noreturn));
static void usageHelp(int const ec) {
  char const * const fmt = "%s - convert GTS data to POV-Ray 'mesh2'.\n"
          "\n  usage: %s [options] < thing.gts > thing.inc\n"
          "\navailable options:\n";
  printf(fmt, prgname, prgname);
  for (size_t i = 0; prgopts_help[i]; ++i)
      printf("  -%s.\n", prgopts_help[i]);
  printf("\n");
  exit(ec);
}

static void usageVersion(void) __attribute__((noreturn));
static void usageVersion(void) {
  char const * const fmt = "%s %s\n"
          "copyright (c) 2020, jr <creature.eternal@gmail.com>.\n"
          "license GPLv3+ <http://gnu.org/licenses/gpl.html>\n"
          "this is free software, you are free to .. whatever.\n"
          "there is no warranty.\n";
  printf(fmt, prgname, prgversion);
  exit(EXIT_SUCCESS);
}

/* ------------------------------------------------------------------------- */

static bool getDouble(char const * const arg, double *val) {
  errno = 0;
  char *p = NULL;
  double d = strtod(arg, &p);
  if (errno || *p) return false;
  *val = d;
  return true;
}

/* min contrib 5% */
static gdouble mkWeight(gdouble const a, gdouble const d) {
  gdouble const w = ((a - arealo) / (d / 100)) / 100;
  return .05 > w ? .05 : w;
}

/* ------------------------------------------------------------------------- */

static void emitPreamble(void) {
  char const * const fmt = "// GTS surface/mesh data.\n"
          "// converted %s (%s v %s).\n\n"
          "#ifndef(%sinclude_temp)\n\n"
          "#declare %sinclude_temp = version;\n"
          "#version %g;\n\n"
          "#ifdef(View_POV_Include_Stack)\n"
          "#debug \"including %s%s.inc\\n\"\n"
          "#end\n\n";
  time_t const now = time(0);
  struct tm *tm = localtime(&now);
  char buf[41] = {CHRNUL};
  strftime(buf, 40, "%d.%m.%Y %H%Mh", tm);

  if (mkbare)
    printf("// converted %s (%s v %s).\n", buf, prgname, prgversion);
  else
    printf(fmt, buf, prgname, prgversion, prefix_value, prefix_value,
            povvers_value, prefix_value, ident_value);
  return;
}

static void emitPostamble(void) {
  char const * const fmt = "\n#version %sinclude_temp;\n\n#end\n\n";
  if (!mkbare)
    printf(fmt, prefix_value);
  return;
}

static void emitAuxiliary(void) {
  printf(auxmask, "begin", aux_value);
  char buf[AUXLNMAX] = {CHRNUL};
  while (fgets(buf, AUXLNMAX, aux_fp))
    if (EOF == fputs(buf, stdout)) EEXIT;
  if (ferror(aux_fp)) EEXIT;
  printf(auxmask, "end", aux_value);
  return;
}

static void cbEmitM1T(GtsFace *fp, guint *n AUU) {
  GtsVertex *v1, *v2, *v3;
  gts_triangle_vertices(&(fp)->triangle, &v1, &v2, &v3);
  printf("  triangle {<%g,%g,%g>, <%g,%g,%g>, <%g,%g,%g>}\n",
          v1->p.x, v1->p.y, v1->p.z,
          v2->p.x, v2->p.y, v2->p.z,
          v3->p.x, v3->p.y, v3->p.z);
  return;
}

static void cbEmitM1ST(GtsFace *fp, guint *n AUU) {
  GtsVertex *v1, *v2, *v3;
  gts_triangle_vertices(&(fp)->triangle, &v1, &v2, &v3);
  size_t const i1 = (size_t) GTS_OBJECT (v1)->reserved,
               i2 = (size_t) GTS_OBJECT (v2)->reserved,
               i3 = (size_t) GTS_OBJECT (v3)->reserved;
  printf("  smooth_triangle {<%g,%g,%g>, <%g,%g,%g>, <%g,%g,%g>, "
        "<%g,%g,%g>, <%g,%g,%g>, <%g,%g,%g>}\n",
          v1->p.x, v1->p.y, v1->p.z,
          vnorms[i1].x, vnorms[i1].y, vnorms[i1].z,
          v2->p.x, v2->p.y, v2->p.z,
          vnorms[i2].x, vnorms[i2].y, vnorms[i2].z,
          v3->p.x, v3->p.y, v3->p.z,
          vnorms[i3].x, vnorms[i3].y, vnorms[i3].z);
  return;
}

static void emitMesh1(GtsSurface *s) {
  printf("#declare %s%s = mesh {\n", prefix_value, ident_value);
  if (mksmooth)
    gts_surface_foreach_face(s, (GtsFunc)cbEmitM1ST, NULL);
  else
    gts_surface_foreach_face(s, (GtsFunc)cbEmitM1T, NULL);
  if (inside) printf(ivecmask, inside_vectors[inside]);
  if (aux_value) emitAuxiliary();
  printf("};\n");
  return;
}

static void cbEmitM2V(GtsVertex *vp, size_t const *n) {
  size_t const i = (size_t) GTS_OBJECT (vp)->reserved;
  printf("    <%g,%g,%g>%s\n",
          vp->p.x, vp->p.y, vp->p.z, *n == i ? "" : ",");
  return;
}

static void cbEmitM2F(GtsFace *fp, size_t const *n) {
  size_t const i = (size_t) GTS_OBJECT (fp)->reserved;
  GtsVertex *v1, *v2, *v3;
  gts_triangle_vertices(&(fp)->triangle, &v1, &v2, &v3);
  /* zero-based indices */
  size_t const i1 = ((size_t) GTS_OBJECT (v1)->reserved) - 1,
               i2 = ((size_t) GTS_OBJECT (v2)->reserved) - 1,
               i3 = ((size_t) GTS_OBJECT (v3)->reserved) - 1;
  printf("    <%zu,%zu,%zu>%s\n", i1, i2, i3, *n == i ? "" : ",");
  return;
}

static void emitMesh2(GtsSurface *s) {
  printf("#declare %s%s = mesh2 {\n", prefix_value, ident_value);
  size_t n = gts_surface_vertex_number(s);
  printf("  vertex_vectors {\n    %zu,\n", n);
  gts_surface_foreach_vertex(s, (GtsFunc)cbEmitM2V, &n);
  printf("  }\n");
  if (mksmooth) {
    printf("  normal_vectors {\n    %zu,\n", n);
    for (size_t i = 1; n >= i; ++i)
      printf("    <%g,%g,%g>%s\n",
              vnorms[i].x, vnorms[i].y, vnorms[i].z,
              n < (i + 1) ? "" : ",");
    printf("  }\n");
  }
  n = gts_surface_face_number(s);
  printf("  face_indices {\n    %zu,\n", n);
  gts_surface_foreach_face(s, (GtsFunc)cbEmitM2F, &n);
  printf("  }\n");
  if (inside) printf(ivecmask, inside_vectors[inside]);
  if (aux_value) emitAuxiliary();
  printf("};\n");
  return;
}

/* callbacks "cribbed" from 'gts2c.c'.
 * not sure why void signatures do not give compiler warnings,
 * (GtsFunc)s are documented as 'gint (*GtsFunc)(..)'.
 * faces - record smallest + largest area + normalized normals.
 */
static void cbMkVertexIndices(GtsVertex *vp, guint *n) {
  GTS_OBJECT (vp)->reserved = GUINT_TO_POINTER((*n)++);
  return;
}

static void cbMkEdgeIndices(GtsSegment *sp, guint *n) {
  GTS_OBJECT (sp)->reserved = GUINT_TO_POINTER((*n)++);
  return;
}

static void cbMkFaceIndices(GtsTriangle *tp, guint *n) {
  gdouble const a = gts_triangle_area(tp);
  if (1 == *n) arealo = areahi = a;
  else if (arealo > a) arealo = a;
  else if (areahi < a) areahi = a;
  gdouble x, y, z;
  gts_triangle_normal(tp, &x, &y, &z);
  GtsVector v;
  v[0] = x, v[1] = y, v[2] = z;
  gts_vector_normalize(v);
  fnorms[*n] = (Vector){v[0], v[1], v[2], true};
  GTS_OBJECT (tp)->reserved = GUINT_TO_POINTER((*n)++);
  return;
}

static void cbMkVnorms(GtsVertex *vp, Cdata const * const dp) {
  size_t const i = (size_t) GTS_OBJECT (vp)->reserved;
  if (!vnorms[i].done) {
    GSList *list = gts_vertex_faces(vp, dp->s, NULL), *p;
    unsigned int n = 0;
    p = list;
    while (p) {
      GtsFace * const fp = p->data;
      size_t const j = (size_t) GTS_OBJECT (fp)->reserved;
      gdouble const w = mkWeight(gts_triangle_area(&fp->triangle), dp->diff);
      vnorms[i].x += fnorms[j].x * w;
      vnorms[i].y += fnorms[j].y * w;
      vnorms[i].z += fnorms[j].z * w;
      n += 1;
      p = p->next;
    }
    vnorms[i].x /= (double)n;
    vnorms[i].y /= (double)n;
    vnorms[i].z /= (double)n;
    vnorms[i].done = true;
    g_slist_free(list);
  }
  return;
}

static int convert(void) {
  bool flag = false;

  GtsSurface *S = gts_surface_new(gts_surface_class(),
                                  gts_face_class(),
                                  gts_edge_class(),
                                  gts_vertex_class());
  GtsFile *file = gts_file_new(stdin);
  if (gts_surface_read(S, file)) {
    fprintf(stderr, "%s: data on <stdin> is not a valid GTS file.\n"
                    "\tat line %d:%d: %s\n",
            prgname, file->line, file->pos, file->error);
    goto label_early_exit;
  }
  else if (diag) {
    GtsBBox const * const b = gts_bbox_surface(gts_bbox_class(), S);
    fprintf(stderr, "# bounding box: <%g, %g, %g>, <%g, %g, %g>\n",
            b->x1, b->y1, b->z1, b->x2, b->y2, b->z2);
    gts_surface_print_stats(S, stderr);
  }
  /* use gts-tools 'cleanup' utility if need repair */
  if (!gts_surface_is_closed(S))
    if (!mkforce) {
      fprintf(stderr, "%s: surface (mesh) is not closed.\n", prgname);
      goto label_early_exit;
    }

  /* one-based serial numbers/indices */
  fnorms = calloc(1 + gts_surface_face_number(S), sizeof(Vector));
  if (!fnorms) EEXIT;
  guint nv, ne, nf;  nv = ne = nf = 1;
  gts_surface_foreach_vertex(S, (GtsFunc)cbMkVertexIndices, &nv);
  if (VERTEXMAX < (nv - 1)) {
    fprintf(stderr, "%s: too many vertices, %d max.\n", prgname, VERTEXMAX);
    goto label_early_exit;
  }
  gts_surface_foreach_edge(S, (GtsFunc)cbMkEdgeIndices, &ne);
  gts_surface_foreach_face(S, (GtsFunc)cbMkFaceIndices, &nf);
  if (mksmooth) {
    vnorms = calloc(1 + gts_surface_vertex_number(S), sizeof(Vector));
    if (!vnorms) EEXIT;
    Cdata x = {.s = S, .diff = areahi - arealo};
    gts_surface_foreach_vertex(S, (GtsFunc)cbMkVnorms, &x);
  }

  emitPreamble();
  if (mkmesh2) emitMesh2(S);
  else emitMesh1(S);
  emitPostamble();
  flag = true;

label_early_exit:
  gts_file_destroy(file);
  gts_finalize();

  return flag ? EXIT_SUCCESS : EXIT_FAILURE;
}

/* ------------------------------------------------------------------------- */

static void argAuxiliary(char const * const arg) {
  aux_fp = fopen(arg, "rb");
  if (!aux_fp) {
    fprintf(stderr, "%s: oops, opening '%s' failed.\n", prgname, arg);
    exit(1);
  }
  aux_value = strdup(arg);
  if (!aux_value) {
    (void)fclose(aux_fp);
    EEXIT;
  }
  return;
}

static bool argInside(char const * const arg) {
  bool rf = true;
  switch (*arg) {
    case 'X': case 'x': inside = kX; break;
    case 'Y': case 'y': inside = kY; break;
    case 'Z': case 'z': inside = kZ; break;
    default:            rf = false;
  }
  return rf;
}

// TODO verify content?
static bool argName(char const * const arg) {
  size_t const n = strlen(arg);
  if (!n || (IDENTMAX <= n))
    return false;
  ident_value = strdup(arg);
  if (!ident_value) EEXIT;
  return true;
}

/* NB. PovRay version magic numbers */
static bool argPovvers(char const * const arg) {
  double d = 0;
  if (!getDouble(arg, &d))
    return false;
  else if (3.1 > d  ||  4.0 <= d)
    return false;
  povvers_value = d;
  return true;
}

// TODO as 'argName'.
static bool argPrefix(char const * const arg) {
  size_t const n = strlen(arg);
  if (!n || (IDENTMAX <= n))
    return false;
  prefix_value = strdup(arg);
  if (!prefix_value) EEXIT;
  return true;
}

static void tidyup(void) {
  if (aux_value) (void)fclose(aux_fp);
  free(aux_value);
  free(ident_value);
  free(prefix_value);
  free(fnorms);
  free(vnorms);
  return;
}

static void setDefaults(void) {
  aux_value = NULL;
  ident_value = NULL;
  inside = kNone;
  mkbare = mkforce = mksmooth = false;
  mkmesh2 = true;
  povvers_value = DFLTVERSION;
  prefix_value = NULL;
  return;
}

int main(int argc, char *argv[]) {
  if (atexit(tidyup)) EEXIT;
  else if (setreuid(getuid(), geteuid())) EEXIT;
  else if (!setlocale(LC_ALL, "POSIX")) EEXIT;
  setDefaults();
  bool have1, havebare, havediag, havefor, haveins,
       havename, havepref, havesmth, havevers;
  int o;
  have1 = havebare = havediag = havefor = haveins
        = havename = havepref = havesmth = havevers = false;
  opterr = 0;
  while (-1 != (o = getopt(argc, argv, prgopts)))
    switch (o) {
      case '1':
        if (have1) {
          fprintf(stderr, "%s: oops, too many 'mesh' arguments.\n", prgname);
          exit(EXIT_FAILURE);
        }
        mkmesh2 = false;
        have1 = true;
        break;
      case 'a':
        if (aux_value) {
          fprintf(stderr, "%s: oops, too many auxiliary files.\n", prgname);
          exit(EXIT_FAILURE);
        }
        argAuxiliary(optarg);
        break;
      case 'b':
        if (havebare) {
          fprintf(stderr, "%s: oops, too many \"bare\" arguments.\n", prgname);
          exit(EXIT_FAILURE);
        }
        mkbare = true;
        havebare = true;
        break;
      case 'D':
        if (havediag) EEXIT;
        diag = true;
        havediag = true;
        break;
      case 'f':
        if (havefor) {
          fprintf(stderr, "%s: oops, too many \"force\" arguments.\n",
                  prgname);
          exit(EXIT_FAILURE);
        }
        mkforce = true;
        havefor = true;
        break;
      case 'h':  usageHelp(0);
      case 'i':
        if (haveins || !argInside(optarg)) {
          fprintf(stderr, "%s: oops, bad \"inside\" argument.\n", prgname);
          exit(EXIT_FAILURE);
        }
        haveins = true;
        break;
      case 'n':
        if (havename || !argName(optarg)) {
          fprintf(stderr, "%s: oops, bad identifier.\n", prgname);
          exit(EXIT_FAILURE);
        }
        havename = true;
        break;
      case 'P':
        if (havevers || !argPovvers(optarg)) {
          fprintf(stderr, "%s: oops, need one \"version\", 3.5 to 3.8.\n",
                  prgname);
          exit(EXIT_FAILURE);
        }
        havevers = true;
        break;
      case 'p':
        if (havepref || !argPrefix(optarg)) {
          fprintf(stderr, "%s: oops, bad identifier prefix.\n", prgname);
          exit(EXIT_FAILURE);
        }
        havepref = true;
        break;
      case 's':
        if (havesmth) {
          fprintf(stderr, "%s: oops, too many \"smooth\" arguments.\n",
                  prgname);
          exit(EXIT_FAILURE);
        }
        mksmooth = true;
        havesmth = true;
        break;
      case 'V': case 'v':  usageVersion();
      default:   usageHelp(2);
    }

  if (argc != optind) {
    fprintf(stderr, "%s: oops, extra argument(s) on command line.\n",
            prgname);
    exit(EXIT_FAILURE);
  }
  /* final preps */
  if (!havename) if (!(ident_value = strdup(DFLTIDENT))) EEXIT;
  if (!havepref) if (!(prefix_value = strdup(DFLTPREFIX))) EEXIT;

  return convert();
}

/* ------------------------------------------------------------- *
 * the content above is covered by the GPLv3+, see file COPYING. *
 * copyright (c) 2020 jr <creature.eternal@gmail.com>.           *
 * all rights reserved.                                          *
 * ------------------------------------------------------------- *
 */
